import urllib
import requests
import logging
import base64
import json


from requests_oauthlib import OAuth2Session


from django.views.generic import View
from django.http import HttpResponseRedirect
from django.contrib.auth import authenticate, login as auth_login, \
    REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import user_passes_test
from django.contrib import messages
from django.shortcuts import resolve_url, render
from django.utils.translation import ugettext as _
from django.utils.http import is_safe_url
from django.conf import settings
from django.core.cache import InvalidCacheBackendError, get_cache


from . import app_settings, models


def user_has_fcaccount(user):
    '''Return True if user a link to FC'''
    try:
        return user.fcaccount is not None
    except models.FcAccount.DoesNotExist:
        return False


fcaccount_required = user_passes_test(user_has_fcaccount, '/')


class LoggerMixin(object):
    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger(__name__)
        super(LoggerMixin, *args, **kwargs)

try:
    cache = get_cache('fc')
except InvalidCacheBackendError:
    cache = get_cache('default')


CACHE_TIMEOUT = 60


def ask_authorization(request, scopes):
    '''Compute an authorize URL for obtaining the given scope'''
    if not isinstance(scopes, (list, tuple)):
        scopes = [scopes]
    redirect_uri = request.build_absolute_uri()
    params = {
        'client_id': app_settings.client_id,
        'scope': ' '.join(scopes),
        'redirect_uri': redirect_uri,
        'response_type': 'code',
        'state': base64.b64encode(redirect_uri),
    }
    url = '{0}?{1}'.format(app_settings.authorize_url,
                           urllib.urlencode(params))
    return HttpResponseRedirect(url)


def resolve_access_token(authorization_code, redirect_uri):
    '''Exchange an authorization_code for an access_token'''
    data = {
        'code': authorization_code,
        'client_id': app_settings.client_id,
        'client_secret': app_settings.client_secret,
        'redirect_uri': redirect_uri,
        'grant_type': 'authorization_code',
    }
    response = requests.post(app_settings.token_url, data=data,
                             verify=app_settings.verify_certificate)
    return response.json()


def access_token_from_request(request):
    '''Resolve an access token given a request returning from the authorization
       endpoint.
    '''
    authorization_code = request.GET['code']
    if 'state' in request.GET:
        state = request.GET.get('state')
        redirect_uri = base64.b64decode(state)
    else:
        redirect_uri = request.GET['redirect_uri']
    return resolve_access_token(authorization_code, redirect_uri)


ACCESS_GRANT_CODE = 'accessgrantcode'


class FcOAuthSessionViewMixin(LoggerMixin):
    '''Add the OAuth2 dance to a view'''
    scopes = ['openid', 'profile']
    redirect_field_name = REDIRECT_FIELD_NAME
    in_popup = False
    token = None

    def get_in_popup(self):
        return self.in_popup

    def redirect_to(self, request, *args, **kwargs):
        redirect_to = request.REQUEST.get(self.redirect_field_name, '')
        if not is_safe_url(url=redirect_to, host=request.get_host()):
            redirect_to = resolve_url(settings.LOGIN_REDIRECT_URL)
        return redirect_to

    def close_popup_redirect(self, request, next_url, *args, **kwargs):
        '''Show a page to close the current popup and reload the parent window
           with the return url.
        '''
        return render(request, 'authentic2_auth_fc/close-popup-redirect.html',
                      {'redirect_to': next_url})

    def simple_redirect(self, request, next_url, *args, **kwargs):
        return HttpResponseRedirect(next_url)

    def redirect(self, request, *args, **kwargs):
        next_url = kwargs.pop('next_url', None)
        if next_url is None:
            next_url = self.redirect_to(request, *args, **kwargs)
        if self.get_in_popup():
            return self.close_popup_redirect(request, next_url, *args,
                                             **kwargs)
        else:
            return self.simple_redirect(request, next_url, *args, **kwargs)

    def redirect_and_come_back(self, request, next_url, *args, **kwargs):
        old_next_url = self.redirect_to(request, *args, **kwargs)
        here = '{0}?{1}'.format(
            request.path,
            urllib.urlencode({REDIRECT_FIELD_NAME: old_next_url}))
        there = '{0}{2}{1}'.format(
            next_url, urllib.urlencode({REDIRECT_FIELD_NAME: here}),
            '&' if '?' in next_url else '?')
        return self.redirect(request, next_url=there, *args, **kwargs)

    def get_scopes(self):
        return self.scopes

    def get_user_info(self):
        return self.oauth_session().get(
            app_settings.userinfo_url + '?schema=openid',
            verify=app_settings.verify_certificate).json()

    def authorization_error(self, request, *args, **kwargs):
        if request.REQUEST['error'] == 'access_denied':
            messages.warning(request, _('request denied by user'))
        self.logger.debug('authorization_error %r', request.GET)
        return self.redirect(request)

    def dispatch(self, request, *args, **kwargs):
        '''Interpret the OAuth authorization dance'''
        if 'code' in request.GET:
            self.token = access_token_from_request(request)
            self.id_token_raw = self.token['id_token']
            # Extract payload from JWT
            self.id_token = json.loads(
                base64.b64_decode(
                    self.id_token_raw.split('.')[1]))
            self.logger.debug('fc id_token %s', self.id_token)
            for key in self.id_token:
                setattr(self, key, self.id_token[key])
            self.oauth_session = lambda: OAuth2Session(
                app_settings.client_id, token=self.token)
            self.user_info = self.get_user_info()
            self.logger.debug('fc user_info %s', self.user_info)
            self.request.session['fc-id_token'] = self.id_token
            self.request.session['fc-user_info'] = self.user_info
            return super(FcOAuthSessionViewMixin, self).dispatch(request,
                                                                 *args,
                                                                 **kwargs)
        elif 'error' in request.GET:
            return self.authorization_error(request, *args, **kwargs)
        else:
            return ask_authorization(request, self.get_scopes())


class PopupViewMixin(object):
    def get_in_popup(self):
        return 'popup' in self.request.REQUEST


class LoginOrLinkView(PopupViewMixin, FcOAuthSessionViewMixin, View):
    '''Login with FC, if the FC account is already linked, connect this user,
       if a user is logged link the user to this account, otherwise display an
       error message.
    '''
    def get(self, request, *args, **kwargs):
        '''Request an access grant code and associate it to the current user'''
        user = authenticate(sub=self.sub, user_info=self.user_info,
                            token=self.token)
        if user:
            auth_login(request, user)
            self.logger.info('logged in using fc sub %s', self.sub)
            return self.redirect(request)
        elif request.user.is_authenticated():
            # Old link are deleted
            json_token = json.dumps(self.token)
            fc_account, created = models.FcAccount.objects.get_or_create(
                defaults={'token': json_token},
                sub=self.sub, user=request.user)

            if created:
                self.logger.info('fc link created sub %s', self.sub)
            else:
                fc_account.token = json_token
                fc_account.save()
            messages.info(request, _('fc link created'))
            return self.redirect(request)
        else:
            messages.info(request, _('to create a link with fc, please '
                                     'authenticate'))
            return self.redirect_and_come_back(
                request, '{0}?nofc=1'.format(settings.LOGIN_URL))


login_or_link = LoginOrLinkView.as_view()
